#ifndef __HS_LZSS_H
#define __HS_LZSS_H

//
// lzss.h
// (c)2002 Jonathan Bennett (jon@hiddensoft.com)
//
// STANDALONE CLASS
//
// This is a memory-to-memory implementation of the LZSS compression
// algorithm.  See the "lzss.txt" file for a hopefully useful explanation
// of how LZSS works.
//
// Note, this is a "as simple to follow as possible" piece of code, there are many
// more efficient and quickler (and cryptic) methods but this was done as a learning
// exercise.
//

// Usage:
// ======
//
// Compress
// --------
// Call the "Compress" function with two caller allocated buffers (unsigned chars)
// The first buffer is the data to compress, the second is where to stored the compressed
// data.  Both buffers should be the same size, if the data cannot be compressed
// an error will be returned (the error will be trapped before the output buffer is overrun,
// so there is no need to supply an output buffer that is larger than the input).
//
// Speeds and compression levels
// Input    Winbits Lenbits  Compressed
// -----    ------- -------  ----------
// 2345984  13,     4        1476082
// 2345984  12,     4        1476738
// 2345984  11,     4        1485692, 31 secs
// 2345984  10,     4        1489546, 20 secs
// 2345984  12,     5        1531966
// 2345984  12,     6        1607990, 60 secs
// 2345984  11,     5         
// 2345984  11,     6         
// 2345984  10,     5         
// 2345984  10,     6         
//
//
// The start of the compression stream will contain "other" items apart from data
// LZSS					4 bytes
// Uncompressed Size	4 bytes (1 long)
// Window bits			1 byte
// Match bits			1 byte
// ...
// Compressed data
// ...
//
// Note: under a windows implementation, unsigned int = unsigned long (4 bytes)
// However, int is not _definitely_ 4 bytes.  This has been taken into account with
// this code and longs have been used whenever 4 bytes is required.
//


// Error codes
#define	HS_LZSS_E_OK			0				// OK
#define	HS_LZSS_E_BADCOMPRESS	1				// Compressed file would be bigger than source!
#define HS_LZSS_E_NOTLZSS		2				// Not a valid LZSS data stream

// Stream flags
#define HS_LZSS_LITERAL			0				// Just output the literal byte
#define HS_LZSS_MATCH			1				// Output a (offset, len) match pair

// Compression options, these will be user selectable in a later version
#define	HS_LZSS_WINDOWLEN		1023			// Sliding window size (10 bits, 0-1023)
#define HS_LZSS_WINDOWBITS		10				// Num of bits that this is
#define HS_LZSS_MATCHLEN		7				// Maximum size of match (3 bits, 0-7)
#define HS_LZSS_MATCHBITS		3				// Num of bits that this is
#define HS_LZSS_MINMATCHLEN		3				// Minimum match size (3 bytes or not efficient)

#define HS_LZSS_MINDATASIZE		32				// Arbitrary miniumum data size that we should attempt to compress
												


class HS_LZSS
{
public:
	// Functions

	int		Compress(unsigned char *bData, unsigned char *bCompressedData, 
						unsigned long nDataSize, unsigned long *nCompressedSize);

	int		Uncompress(unsigned char *bCompressedData, unsigned char *bData);

	int		GetUncompressedSize(unsigned char *bCompressedData, unsigned long *nUncompressedSize);



private:
	// Variables
	unsigned long	m_nDataStreamPos;					// Current position in the data stream
	unsigned long	m_nCompressedStreamPos;				// Curent position in the compressed stream
	unsigned char	*m_bData;
	unsigned char	*m_bCompressedData;

	unsigned long	m_nDataSize;						// The size of our uncompressed data
	unsigned long	m_nCompressedSize;					// The size of our compressed data

	// Temporary variables used for the bit operations
	unsigned long	m_nCompressedLong;					// Compressed stream temporary 32bit value
	int				m_nCompressedBitsUsed;				// Number of bits used in temporary value


	// Functions
	bool	FindMatches(unsigned int *nOffset, unsigned int *nLen);		// Searches for pattern matches

	// Bit operation functions
	void	CompressedStreamWriteBits(unsigned int nValue, unsigned int nNumBits);
	void	CompressedStreamReadBits(unsigned int *nValue, unsigned int nNumBits);
};


#endif
