//
// lzss_compress.cpp
// (c)2002 Jonathan Bennett (jon@hiddensoft.com)
//
//

#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "..\include\lzss.h"


///////////////////////////////////////////////////////////////////////////////
// Compress()
///////////////////////////////////////////////////////////////////////////////

int HS_LZSS::Compress(unsigned char *bData, unsigned char *bCompressedData, 
						unsigned long nDataSize, unsigned long *nCompressedSize, 
						unsigned int nWindowBits, unsigned int nLenBits)
{
	unsigned int	nOffset, nLen;

	// Set up initial values
	m_nDataStreamPos		= 0;				// We are at the start of the input data
	m_nCompressedStreamPos	= 0;				// We are at the start of the compressed data
	m_nCompressedLong		= 0;				// Compressed stream temporary 32bit value
	m_nCompressedBitsUsed	= 0;				// Number of bits used in temporary value
	m_bCompressedData		= bCompressedData;	// Pointer to our output buffer
	m_bData					= bData;			// Pointer to our input buffer
	m_nDataSize				= nDataSize;		// Store the size of our input data
	m_nCompressedSize		= 0;				// No compressed data yet


	// If the input file is too small then there is a chance of 
	// buffer overrun, so just abort
	if (m_nDataSize < HS_LZSS_MINDATASIZE)
		return HS_LZSS_E_BADCOMPRESS;

	// Setup our user defined window size and length match size
	SetupWindowSize(nWindowBits, nLenBits);

	// Write out our file header info so the decompressor doesn't have to
	// guess about things like uncompressed buffer sizes
	// LZSS ID (4 bytes)
	CompressedStreamWriteBits('L', 8);	CompressedStreamWriteBits('Z', 8);
	CompressedStreamWriteBits('S', 8);	CompressedStreamWriteBits('S', 8);

	// Uncompressed size (4 bytes)
	CompressedStreamWriteBits( (m_nDataSize >> 16) & 0x0000ffff, 16);
	CompressedStreamWriteBits( m_nDataSize & 0x0000ffff, 16);

	// Window and match bits (not yet implemented)
	CompressedStreamWriteBits( m_nWindowBits, 8);
	CompressedStreamWriteBits( m_nMatchLenBits, 8);


	// Loop around until there is no more data
	while (m_nDataStreamPos < m_nDataSize)
	{
		// If at any point, we reach within 16 bytes of the end of our output buffer, abort!
		// See the end of this while loop for details of what is output after we have
		// finished compressing
		if ( m_nCompressedSize >= (m_nDataSize - 16) )
			return HS_LZSS_E_BADCOMPRESS;

		// Search for matches
		if ( FindMatches(&nOffset, &nLen) == true )
		{
			// A match was found, write it out
			CompressedStreamWriteBits(HS_LZSS_MATCH, 1);
			CompressedStreamWriteBits(nOffset - m_nWindowLenAdjust, m_nWindowBits);
			CompressedStreamWriteBits(nLen - m_nMatchLenAdjust, m_nMatchLenBits);
		} 
		else
		{
			// No matches were found, just write out the literal byte
			CompressedStreamWriteBits(HS_LZSS_LITERAL, 1);
			CompressedStreamWriteBits(m_bData[m_nDataStreamPos++], 8);

		} // End If

	} // End while

	// Write out 2 bytes (16 bits) of nothing to help avoid overruns when decoding
	// Lazy....I know.
	CompressedStreamWriteBits(0, 16);

	// Flush the compressed bitstream (pad with zeros if required)
	CompressedStreamWriteBits(0, 0);

	// Return the compressed data size
	*nCompressedSize = m_nCompressedSize;

	return HS_LZSS_E_OK;						// Return with success message

} // Compress()


///////////////////////////////////////////////////////////////////////////////
// SetupWindowSize()
//
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS::SetupWindowSize(unsigned int nWindowBits, unsigned int nLenBits)
{
	unsigned int nTemp;

	m_nWindowBits			= nWindowBits;
	m_nMatchLenBits			= nLenBits;
	m_nMatchLenAdjust		= 0;

	// Setup initial window and bit lens (may be adjusted later for efficiency)
	m_nWindowLen	= ((unsigned int)pow(2, (double)m_nWindowBits)) - 1;
	m_nMatchLen		= ((unsigned int)pow(2, (double)m_nMatchLenBits)) - 1;

	// How many bits does a match offset take up?
	nTemp = m_nWindowBits + m_nMatchLenBits;

	// Each literal byte takes up 9 bits when encoded, what is the "break even"?
	m_nMinMatchLen = (nTemp / 9) + 1;

	// Based on this information we adjust our lengths
	m_nWindowLenAdjust	= m_nMinMatchLen - 1;
	m_nWindowLen		= m_nWindowLen + m_nWindowLenAdjust;
	m_nMatchLenAdjust	= m_nMinMatchLen - 1;
	m_nMatchLen			= m_nMatchLen + m_nMatchLenAdjust;

	printf("Number of window bits: %d (0-%d)\n", m_nWindowBits, m_nWindowLen);
	printf("Number of length bits: %d (0-%d)\n", m_nMatchLenBits, m_nMatchLen);
	printf("Minimum match len    : %d\n", m_nMinMatchLen);
	printf("Window len adjust    : %d\n", m_nWindowLenAdjust);
	printf("Match len adjust     : %d\n", m_nMatchLenAdjust);

} // SetupWindowSize()


///////////////////////////////////////////////////////////////////////////////
// CompressedStreamWriteBits()
//
// Will write a number of bits (variable) into the compressed data stream
//
// When there are no more bits to send, you should call the function with the
// parameters 0, 0 to make sure that any left over bits are flushed into the
// compressed stream
//
// Note 16 bits is the maximum allowed value for this function!!!!!
// ================================================================
// This equates to a maximum window size of 65535
//
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS::CompressedStreamWriteBits(unsigned int nValue, unsigned int nNumBits)
{
	if (nNumBits == 0)
	{
		// We have been asked to finish, flush remaining by using up remaining
		// unused bits with zeros

		// Bits remaining = 32 minus bits used
		m_nCompressedLong = m_nCompressedLong << (32 - m_nCompressedBitsUsed);

		// Now write out all 32 bits
		m_bCompressedData[m_nCompressedStreamPos++] = (unsigned char)(m_nCompressedLong >> 24);
		m_bCompressedData[m_nCompressedStreamPos++] = (unsigned char)(m_nCompressedLong >> 16);
		m_bCompressedData[m_nCompressedStreamPos++] = (unsigned char)(m_nCompressedLong >> 8);
		m_bCompressedData[m_nCompressedStreamPos++] = (unsigned char)(m_nCompressedLong);

		// We have just stored 4 bytes of data, increase our size count
		m_nCompressedSize = m_nCompressedSize + 4;

		// All done
		m_nCompressedBitsUsed = 0;

		return;
	}

	while (nNumBits > 0)
	{
		nNumBits--;

		// Make room for another bit (shift left once)
		m_nCompressedLong = m_nCompressedLong << 1;

		// Merge (OR) our value into the temporary long
		m_nCompressedLong = m_nCompressedLong | (long)((nValue >> nNumBits) & 0x00000001);
		//printf("%d ", ((nValue >> nNumBits) & 0x00000001));

		// Update how many bits we are using (add 1)
		m_nCompressedBitsUsed++;

		// Now check if we have filled our temporary long with bits (32bits)
		if (m_nCompressedBitsUsed == 32)
		{
			// We now need to dump the highest 16 bits into our compressed
			// stream.  Highest order 8 bits first
			m_bCompressedData[m_nCompressedStreamPos++] = (unsigned char)(m_nCompressedLong >> 24);
			m_bCompressedData[m_nCompressedStreamPos++] = (unsigned char)(m_nCompressedLong >> 16);

			// We have just stored 2 bytes of data, increase our size count
			m_nCompressedSize = m_nCompressedSize + 2;

			// We've now written out 16 bits so make more room (16 bits more room :) )
			m_nCompressedBitsUsed = m_nCompressedBitsUsed - 16;
		}

	} // End while


} // CompressedStreamWriteBits()


///////////////////////////////////////////////////////////////////////////////
// FindMatches()
//
// OK, this is the "money" function.  From our current position in the Data stream
// go back in the data stream by "window" size.  Then look for the longest match
// with the current position in the data stream...Er.....yeah....I understood that...
//
// This function is reallllllly slow.  
//  2345984 bytes compressed in 65 seconds (366MHz CPU). :)
//
// Using member variables from the class appears to really slow things down, so 
// Local versions are created to speed things up (in somecases)
//
// bData[nDataStreamPos] == bData[nWPos] 
// is quicker than
// cTempChar == bData[nWPos]  !!!!!
//
///////////////////////////////////////////////////////////////////////////////

bool HS_LZSS::FindMatches(unsigned int *nOffset, unsigned int *nLen)
{
	unsigned long	nTempWPos, nWPos, nDPos;	// Temp Window and Data position markers
	unsigned long	nMaxWPos, nDataStreamPos, nDataSize;	// Pure temp values for speeding up loops
	unsigned long	nTempLen;					// Temp vars 
	unsigned long	nBestOffset, nBestLen;		// Stores the best match so far
	unsigned char	*bData;

	// Reset all variables
	nBestOffset = 0;
	nBestLen	= m_nMinMatchLen - 1;

	// If the current data stream position is less than the minimum match length bytes
	// into the stream then we definately won't match so just quit
	if (m_nDataStreamPos < m_nMinMatchLen)
		return false;							// No matches


	// Get our window start position, if the window would take us beyond
	// the start of the file, just use 0
	if (m_nDataStreamPos < m_nWindowLen)
		nWPos = 0;
	else
		nWPos = m_nDataStreamPos - m_nWindowLen;


	// Store some temp values to speed up the loop (a little)
	nDataStreamPos	= m_nDataStreamPos;
	nMaxWPos		= nDataStreamPos - m_nMinMatchLen;
	bData			= m_bData;
	nDataSize		= m_nDataSize;
	

	// Main loop
	while ( nWPos <= nMaxWPos)
	{
		if ( bData[nDataStreamPos] == bData[nWPos] )
		{
			// Found an initial match, lets see how long it is
			nTempWPos	= nWPos;
			nDPos		= nDataStreamPos;
			nTempLen	= 0;

			while ( (bData[nTempWPos] == bData[nDPos]) && (nTempWPos < nDataStreamPos) &&
					nDPos < nDataSize) 
			{
				nTempLen++;
				nTempWPos++;
				nDPos++;
			} // End while

			// See if this match was better than previous match
			if (nTempLen > nBestLen)
			{
				if (nTempLen >= m_nMatchLen)
				{
					nBestLen	= m_nMatchLen;
					nBestOffset = nDataStreamPos - nWPos;
					nWPos		= nMaxWPos;		// Force the while loop to end
				}
				else
				{
					nBestLen	= nTempLen;
					nBestOffset = nDataStreamPos - nWPos;
				}
			}
		} // End if

		nWPos++;

	} // End while


	// If we matched, then return the best values
	if ( (nBestOffset == 0) )
		return false;							// No match
	else
	{
		// Move our data stream pointer further on in the file
		m_nDataStreamPos = m_nDataStreamPos + nBestLen;

		// Return the best values
		*nOffset	= nBestOffset;
		*nLen		= nBestLen;
		return true;							// Match! :)
	}

} // FindMatches()

