//
// lzss_uncompress.cpp
// (c)2002 Jonathan Bennett (jon@hiddensoft.com)
//
//

#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "..\include\lzss.h"


///////////////////////////////////////////////////////////////////////////////
// CompressedStreamReadBits()
//
// Will read up to 16 bits from the compressed data stream
//
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS::CompressedStreamReadBits(UINT *nValue, UINT nNumBits)
{

	ULONG	nTemp;

	// Ensure that the high order word of our bit buffer is blank 
	m_nCompressedLong = m_nCompressedLong & 0x000ffff;

	while (nNumBits > 0)
	{
		nNumBits--;

		// Check if we need to refill our decoding bit buffer
		if (m_nCompressedBitsUsed == 0)
		{
			// Fill the low order 16 bits of our long buffer
			nTemp = (ULONG)m_bCompressedData[m_nCompressedStreamPos++];
			m_nCompressedLong = m_nCompressedLong | (nTemp << 8);
			nTemp = (ULONG)m_bCompressedData[m_nCompressedStreamPos++];
			m_nCompressedLong = m_nCompressedLong | nTemp;
			m_nCompressedBitsUsed = 16;
		}

		// Shift the data into the high part of the long
		m_nCompressedLong = m_nCompressedLong << 1;
		m_nCompressedBitsUsed--;
	}

	*nValue = (UINT)(m_nCompressedLong >> 16);

} // CompressedStreamReadBits()


///////////////////////////////////////////////////////////////////////////////
// GetUncompressedSize()
///////////////////////////////////////////////////////////////////////////////

int HS_LZSS::GetUncompressedSize(UCHAR *bCompressedData, ULONG *nUncompressedSize)
{
	UINT	nTemp;
	char			szAlgID[4+1];

	// Set up initial values
	m_nCompressedStreamPos	= 0;				// We are at the start of the compressed data
	m_nCompressedLong		= 0;				// Compressed stream temporary 32bit value
	m_nCompressedBitsUsed	= 0;				// Number of bits used in temporary value
	m_bCompressedData		= bCompressedData;	// Pointer to our input buffer


	// Check that data is a valid LZSS stream and return the length of it
	CompressedStreamReadBits(&nTemp, 8);	szAlgID[0] = (char)nTemp;
	CompressedStreamReadBits(&nTemp, 8);	szAlgID[1] = (char)nTemp;
	CompressedStreamReadBits(&nTemp, 8);	szAlgID[2] = (char)nTemp;
	CompressedStreamReadBits(&nTemp, 8);	szAlgID[3] = (char)nTemp;
	szAlgID[4] = '\0';

	if ( stricmp(szAlgID, "LZSS") )
		return HS_LZSS_E_NOTLZSS;				// Wasn't a valid LZSS stream

	// Uncompressed size (4 bytes)
	CompressedStreamReadBits( &nTemp, 16); 
	*nUncompressedSize = ((ULONG)nTemp) << 16;
	CompressedStreamReadBits( &nTemp, 16);
	*nUncompressedSize = *nUncompressedSize | (ULONG)nTemp;

	return HS_LZSS_E_OK;

} // GetUncompressedSize()


///////////////////////////////////////////////////////////////////////////////
// Uncompress()
//
// This function ASSUMES that our decompression buffer is big enough, i.e.
// The caller should have done GetUncompressedSize() first and allocated 
// a buffer big enough!
//
// Also, the GetUncompressedSize checks that the LZSS alg ID was found
// so we won't bother this time
//
// In other words, don't run this function on random areas of memory, it will
// crash.
//
///////////////////////////////////////////////////////////////////////////////

int HS_LZSS::Uncompress(UCHAR *bCompressedData, UCHAR *bData)
{
	UINT	nTemp;
	UINT	nOffset, nLen;
	ULONG	nTempPos;
	UINT	nWinBits, nMatchBits;

	// Set up initial values
	m_nDataStreamPos		= 0;				// We are at the start of the input data
	m_nCompressedStreamPos	= 0;				// We are at the start of the compressed data
	m_nCompressedLong		= 0;				// Compressed stream temporary 32bit value
	m_nCompressedBitsUsed	= 0;				// Number of bits used in temporary value
	m_bCompressedData		= bCompressedData;	// Pointer to our input buffer
	m_bData					= bData;			// Pointer to our output buffer
	m_nDataSize				= 0;				// We will get this from the input


	// Skip the LZSS alg ID (4 bytes)
	CompressedStreamReadBits(&nTemp, 16);	CompressedStreamReadBits(&nTemp, 16);	

	// Get the uncompressed size (4 bytes)
	CompressedStreamReadBits( &nTemp, 16); 	m_nDataSize = ((ULONG)nTemp) << 16;
	CompressedStreamReadBits( &nTemp, 16);	m_nDataSize = m_nDataSize | (ULONG)nTemp;

	// Get the window and bit lengths (2 bytes)
	CompressedStreamReadBits(&nWinBits, 8);	CompressedStreamReadBits(&nMatchBits, 8);	
	SetupWindowSize(nWinBits, nMatchBits);


	// Perform decompression until we fill our predicted buffer
	while(m_nDataStreamPos < m_nDataSize)
	{
		// Read in the 1 bit flag
		CompressedStreamReadBits(&nTemp, 1);	

		// Was it a literal byte, or a (offset,len) match pair?
		if (nTemp == HS_LZSS_MATCH)
		{
			// Read the offset and length
			CompressedStreamReadBits(&nOffset, m_nWindowBits);
			CompressedStreamReadBits(&nLen, m_nMatchLenBits);
			
			// Adjust the values
			nOffset += m_nWindowLenAdjust;;
			nLen	+= m_nMatchLenAdjust;


			// Write out our match
			nTempPos = m_nDataStreamPos - nOffset;
			while (nLen > 0)
			{
				nLen--;
				m_bData[m_nDataStreamPos++] = m_bData[nTempPos++];	
			}
		}
		else
		{
			// Output a literal byte
			CompressedStreamReadBits(&nTemp, 8);
			m_bData[m_nDataStreamPos++] = (UCHAR)nTemp;
		}
	}
	
	return HS_LZSS_E_OK;

} // Uncompress()

