#ifndef __HS_LZSS_COMPRESS_H
#define __HS_LZSS_COMPRESS_H

//
// lzss_compress.h
// (c)2002 Jonathan Bennett (jon@hiddensoft.com)
//
// STANDALONE CLASS
//
// This is a memory-to-memory implementation of the LZSS compression
// algorithm.  See the "lzss.txt" file for a hopefully useful explanation
// of how LZSS works.
//
// Note, this is a "as simple to follow as possible" piece of code, there are many
// more efficient and quicker (and cryptic) methods but this was done as a learning
// exercise.
//

// Usage
// =====
// Call "Compress()" with two UCHAR buffers, one for the data to compress and one for
// the compressed output, the size of the input biffer and the required compression level.  
// >> The output buffer should be of the same size as the input.<<
// If the compressed data would have overun the output buffer, or the data is too small
// to compress the function will abort and return the code HS_LZSS_E_BADCOMPRESS.  If the
// function is successful it will return HS_LZSS_E_OK and the size of the new compressed data
// in "nCompressedSize".
//
// The compression level parameter (0-4) gives the following results:
// Compression Level  Window Bits  Match Bits  Hash Limit
// -----------------  -----------  ----------  ----------
// Fast       (0)     12           4           2               
// Normal     (1)     13           4           8
// Slow       (2)     13           4           16
// Very slow  (3)     14           4           32
// Ultra slow (4)     14           4           128
//

//
// The start of the compression stream will contain "other" items apart from data
// LZSS					4 bytes
// Uncompressed Size	4 bytes (1 ULONG)
// Window bits			1 byte
// Match bits			1 byte
// ...
// Compressed data
// ...
//


// Error codes
#define	HS_LZSS_E_OK			0				// OK
#define	HS_LZSS_E_BADCOMPRESS	1				// Compressed file would be bigger than source!
#define HS_LZSS_E_NOTLZSS		2				// Not a valid LZSS data stream

// Stream flags
#define HS_LZSS_LITERAL			0				// Just output the literal byte
#define HS_LZSS_MATCH			1				// Output a (offset, len) match pair

#define HS_LZSS_MINDATASIZE		32				// Arbitrary miniumum data size that we should attempt to compress

// Define our hash and hash table structure
#define HS_LZSS_HASHTYPE1						// Define this to use hash function 1

#ifdef HS_LZSS_HASHTYPE1
	#define LZSS_HASHTABLE_SIZE		4096		// Our hash function gives values 0-4095
#else
	#define LZSS_HASHTABLE_SIZE		65536		// Our hash function gives values 0-65535
#endif


typedef struct LZSS_Hash
{
	ULONG	nPos;								// Position in data stream
	struct 	LZSS_Hash *lpNext;					// Next entry in linked list

} _LZSS_Hash;

												


class HS_LZSS_Compress
{
public:
	// Functions

	int		Compress(UCHAR *bData, UCHAR *bCompressedData, 
						ULONG nDataSize, ULONG *nCompressedSize,
						UINT nCompressionLevel);

private:
	// Variables
	ULONG	m_nDataStreamPos;					// Current position in the data stream
	ULONG	m_nCompressedStreamPos;				// Curent position in the compressed stream
	UCHAR	*m_bData;
	UCHAR	*m_bCompressedData;

	ULONG	m_nDataSize;						// The size of our uncompressed data
	ULONG	m_nCompressedSize;					// The size of our compressed data

	// Temporary variables used for the bit operations
	ULONG	m_nCompressedLong;					// Compressed stream temporary 32bit value
	int		m_nCompressedBitsUsed;				// Number of bits used in temporary value

	// Window size related variables
	UINT	m_nWindowBits;						// Bits used for window size
	UINT	m_nWindowLen;						// Window range
	UINT	m_nMatchLenBits;					// Bits used for match len size
	UINT	m_nMatchLen;						// Match len range
	UINT	m_nMinMatchLen;						// Minimum match len

	// Hash table
	UINT	m_nHashChainLimit;					// The max length of each hash chain
	struct	LZSS_Hash *m_HashTable[LZSS_HASHTABLE_SIZE];	// Hash table 

	// Functions
	int		CompressLoop(void);					// The main compression loop
	void	SetupWindowSize(UINT nCompressionLevel);	// Init the window and len sizes
	void	FindMatches(UINT *nOffset, UINT *nLen);		// Searches for pattern matches

	// Bit operation functions
	void	CompressedStreamWriteBits(UINT nValue, UINT nNumBits);

	// Hash table functions
	void	HashTableInit(void);					// Make hash table ready for first use
	void	HashTableFree(void);					// Free all hash entries
	void	HashTableAdd(UINT nHash, ULONG nPos);	// Add an entry to the table
	void	HashTablePrune(UINT nHash, ULONG nOldestPos);	// Remove old entries
};


#endif
