//
// lzss_compress.cpp
// (c)2002 Jonathan Bennett (jon@hiddensoft.com)
//
//

#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "..\include\lzss_compress.h"


///////////////////////////////////////////////////////////////////////////////
// SetupWindowSize()
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS_Compress::SetupCompressionLevel(UINT nCompressionLevel)
{

	switch (nCompressionLevel)
	{
		default:
			m_nHashChainLimit	= 1;	// Cannot be 0.  1 = 2 hash entries
			break;
		case 1:
			m_nHashChainLimit	= 4;
			break;
		case 2:
			m_nHashChainLimit	= 16;
			break;
		case 3:
			m_nHashChainLimit	= 64;
			break;
		case 4:
			m_nHashChainLimit	= 256;
			break;
	}

} // SetupCompressionLevel()


///////////////////////////////////////////////////////////////////////////////
// Compress()
///////////////////////////////////////////////////////////////////////////////

int HS_LZSS_Compress::Compress(UCHAR *bData, UCHAR *bCompressedData, 
						ULONG nDataSize, ULONG *nCompressedSize, 
						UINT nCompressionLevel)
{
	int	nRes;

	// Set up initial values
	m_nDataStreamPos		= 0;				// We are at the start of the input data
	m_nCompressedStreamPos	= 0;				// We are at the start of the compressed data
	m_nCompressedLong		= 0;				// Compressed stream temporary 32bit value
	m_nCompressedBitsUsed	= 0;				// Number of bits used in temporary value
	m_bCompressedData		= bCompressedData;	// Pointer to our output buffer
	m_bData					= bData;			// Pointer to our input buffer
	m_nDataSize				= nDataSize;		// Store the size of our input data
	m_nCompressedSize		= 0;				// No compressed data yet


	// If the input file is too small then there is a chance of 
	// buffer overrun, so just abort
	if (m_nDataSize < HS_LZSS_MINDATASIZE)
		return HS_LZSS_E_BADCOMPRESS;

	// Setup our user defined window size and length match size
	SetupCompressionLevel(nCompressionLevel);

	// Write out our file header info so the decompressor doesn't have to
	// guess about things like uncompressed buffer sizes
	// LZSS ID (4 bytes)
	CompressedStreamWriteBits('J', 8);	
	CompressedStreamWriteBits('B', 8);
	CompressedStreamWriteBits('0', 8);	
	CompressedStreamWriteBits('1', 8);

	// Uncompressed size (4 bytes)
	CompressedStreamWriteBits( (m_nDataSize >> 16) & 0x0000ffff, 16);
	CompressedStreamWriteBits( m_nDataSize & 0x0000ffff, 16);

	// Initialize our hash table
	HashTableInit();

	//
	// Jump to our main compression function
	//

	nRes = CompressLoop();
	if (nRes != HS_LZSS_E_OK)
		return nRes;							// Return error code


	// Write out 2 bytes (16 bits) of nothing to help avoid overruns when decoding
	// Lazy....I know.
	CompressedStreamWriteBits(0, 16);

	// Flush the compressed bitstream (pad with zeros if required)
	CompressedStreamWriteBitsFlush();

	// Free up our hash table
	HashTableFree();

	// Return the compressed data size
	*nCompressedSize = m_nCompressedSize;

	return HS_LZSS_E_OK;						// Return with success message

} // Compress()


///////////////////////////////////////////////////////////////////////////////
// CompressLoop()
//
///////////////////////////////////////////////////////////////////////////////

int HS_LZSS_Compress::CompressLoop(void)
{
	UINT	nOffset1, nLen1;					// n Offset values
	UINT	nOffset2, nLen2;					// n+1 Offset values (lazy evaluation)
	ULONG	nMaxData;							// The largest size we want for compressed data
	ULONG	nDataPos1;							// N data pos temp values
	ULONG	nTempDataPos;
	UINT	nHash;
	UINT	nTempUINT;
	UINT	nWPos;

	nMaxData = m_nDataSize - 16;				// Abort if compressed data is this big

	// Loop around until there is no more data
	while (m_nDataStreamPos < m_nDataSize)
	{
		// If at any point, we reach within 16 bytes of the end of our output buffer, abort!
		// See the Compress() function for details of what is output after we have finished
		if ( m_nCompressedSize >= nMaxData )
			return HS_LZSS_E_BADCOMPRESS;


		// Store where we are before we go crazy with lazy evaluation stuff
		nTempDataPos = m_nDataStreamPos;

		// Match 
		FindMatches(&nOffset1, &nLen1);			// Search for matches for current position
		nDataPos1 = m_nDataStreamPos;			// Store the data pos if we use this match
		
		// Match + 1
		m_nDataStreamPos = nTempDataPos + 1;	// Reset the stream pointer for a match at +1

		// Check that we are not at the end of the data for doing a match at +1
		if (m_nDataStreamPos < m_nDataSize)
			FindMatches(&nOffset2, &nLen2);			// Search for matches for current position +1
		else
		{
			// We will overrun the end of the data buffer by doing a second match
			// fool the routine into thinking that 2nd match was blank
			nOffset2	= 0;
			nLen2		= 0;
		}
		
		
		// Are there ANY good matches?
		if ( (nOffset1 != 0) || (nOffset2 != 0) )
		{
			// Which was the best match? N or N+1?
			if (nLen1 >= nLen2)
			{
				// Let's use match 1
				m_nDataStreamPos = nDataPos1;	// Restore the data pos for this match
				CompressedStreamWriteBits(HS_LZSS_MATCH, 1);
				CompressedStreamWriteBits(nOffset1 - HS_LZSS_MINMATCHLEN, HS_LZSS_WINBITS);
				CompressedStreamWriteBits(nLen1 - HS_LZSS_MINMATCHLEN, HS_LZSS_MATCHBITS);
			}
			else
			{
				// Remember m_nDataStreamPos WILL ALREADY BE VALID for match +1
				// Let's use match 2 (a little more work required)
				// First, store the byte at N as a literal
				CompressedStreamWriteBits(HS_LZSS_LITERAL, 1);
				CompressedStreamWriteBits(m_bData[nTempDataPos], 8);

				// Then store match 2
				//m_nDataStreamPos = nDataPos2;	// Restore the data pos for this match
				CompressedStreamWriteBits(HS_LZSS_MATCH, 1);
				CompressedStreamWriteBits(nOffset2 - HS_LZSS_MINMATCHLEN, HS_LZSS_WINBITS);
				CompressedStreamWriteBits(nLen2 - HS_LZSS_MINMATCHLEN, HS_LZSS_MATCHBITS);
			}
		}
		else
		{
			// No matches, just store the literal byte
			m_nDataStreamPos = nTempDataPos;
			CompressedStreamWriteBits(HS_LZSS_LITERAL, 1);
			CompressedStreamWriteBits(m_bData[m_nDataStreamPos++], 8);
		}

		// We have skipped forwards either 1 byte or xxx bytes (if matched) we must now
		// add entries in the hash table for all the entries we've skipped

		if (nTempDataPos < HS_LZSS_WINLEN)
			nWPos = 0;
		else
			nWPos = nTempDataPos - HS_LZSS_WINLEN;
		
		nTempUINT = m_nDataStreamPos - nTempDataPos;	// How many bytes to hash?
		while (nTempUINT > 0)
		{
			nTempUINT--;
			nHash =((40543*((((m_bData[nTempDataPos]<<4)^m_bData[nTempDataPos+1])<<4)^m_bData[nTempDataPos+2]))>>4) & 0xFFF;
			HashTableAdd(nHash, nTempDataPos, nWPos);	
			nTempDataPos++;
			
		}  // End while

	} // End while

	return HS_LZSS_E_OK;						// Return with success message

} // CompressLoop()


///////////////////////////////////////////////////////////////////////////////
// CompressedStreamWriteBits()
//
// Will write a number of bits (variable) into the compressed data stream
//
// When there are no more bits to send, you should call the function with the
// parameters 0, 0 to make sure that any left over bits are flushed into the
// compressed stream
//
// Note 16 bits is the maximum allowed value for this function!!!!!
// ================================================================
// This equates to a maximum window size of 65535
//
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS_Compress::CompressedStreamWriteBits(UINT nValue, UINT nNumBits)
{
	while (nNumBits > 0)
	{
		nNumBits--;

		// Make room for another bit (shift left once)
		m_nCompressedLong = m_nCompressedLong << 1;

		// Merge (OR) our value into the temporary long
		m_nCompressedLong = m_nCompressedLong | ((nValue >> nNumBits) & 0x00000001);

		// Update how many bits we are using (add 1)
		m_nCompressedBitsUsed++;

		// Now check if we have filled our temporary long with bits (32bits)
		if (m_nCompressedBitsUsed == 32)
		{
			// We now need to dump the highest 16 bits into our compressed
			// stream.  Highest order 8 bits first
			m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong >> 24);
			m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong >> 16);

			// We have just stored 2 bytes of data, increase our size count
			m_nCompressedSize = m_nCompressedSize + 2;

			// We've now written out 16 bits so make more room (16 bits more room :) )
			m_nCompressedBitsUsed = m_nCompressedBitsUsed - 16;
		}

	} // End while


} // CompressedStreamWriteBits()


///////////////////////////////////////////////////////////////////////////////
// CompressedStreamWriteBitsFlush()
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS_Compress::CompressedStreamWriteBitsFlush(void)
{
	// We have been asked to finish, flush remaining by using up remaining
	// unused bits with zeros

	// Bits remaining = 32 minus bits used
	m_nCompressedLong = m_nCompressedLong << (32 - m_nCompressedBitsUsed);

	// Now write out all 32 bits
	m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong >> 24);
	m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong >> 16);
	m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong >> 8);
	m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong);

	// We have just stored 4 bytes of data, increase our size count
	m_nCompressedSize = m_nCompressedSize + 4;

	// All done
	m_nCompressedBitsUsed = 0;

} // CompressedStreamWriteBitsFlush()


///////////////////////////////////////////////////////////////////////////////
// FindMatches()
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS_Compress::FindMatches(UINT *nOffset, UINT *nLen)
{
	ULONG	nTempWPos, nWPos, nDPos;	// Temp Window and Data position markers
	ULONG	nTempLen;					// Temp vars 
	ULONG	nBestOffset, nBestLen;		// Stores the best match so far
	struct LZSS_Hash *lpTempHash, *lpTempPrev, *lpTempNext;
	UINT	nHash;

	// Reset all variables
	nBestOffset = 0;
	nBestLen	= HS_LZSS_MINMATCHLEN - 1;

	// Get our window start position, if the window would take us beyond
	// the start of the file, just use 0
	if (m_nDataStreamPos < HS_LZSS_WINLEN)
		nWPos = 0;
	else
		nWPos = m_nDataStreamPos - HS_LZSS_WINLEN;

	// Generate a hash of the next three chars
	nHash =((40543*((((m_bData[m_nDataStreamPos]<<4)^m_bData[m_nDataStreamPos+1])<<4)^m_bData[m_nDataStreamPos+2]))>>4) & 0xFFF;

	// Main loop
	lpTempPrev = NULL;
	lpTempHash = m_HashTable[nHash];			// Get our match from the hash table

	while (lpTempHash != NULL)
	{
		nTempWPos	= lpTempHash->nPos;			// The position of the supposed match
		lpTempNext	= lpTempHash->lpNext;		// Next hash

		// Is this entry too old to be used for a match?
		if (nTempWPos < nWPos)
		{
			// Remove it!
			if (lpTempPrev != NULL)
				lpTempPrev->lpNext = lpTempNext;
			else
				m_HashTable[nHash] = lpTempNext;

			//free(lpTempHash);
			HashEntryFree(lpTempHash);

			lpTempHash = lpTempNext;
		}
		else
		{
			nDPos		= m_nDataStreamPos;
			nTempLen	= 0;
		
			while ( (m_bData[nTempWPos] == m_bData[nDPos]) && (nTempWPos < m_nDataStreamPos) &&
					nDPos < m_nDataSize && (nTempLen < HS_LZSS_MATCHLEN) ) 
			{
				nTempLen++;	nTempWPos++; nDPos++;
			} // End while

			// See if this match was better than previous match
			if (nTempLen > nBestLen)
			{
				nBestLen	= nTempLen;
				nBestOffset = m_nDataStreamPos - lpTempHash->nPos;
			}

			lpTempPrev = lpTempHash;
			lpTempHash = lpTempNext;

		}// End if
	} // End while


	// Setup our return values of bestoffset and bestlen, bestoffset will be 0
	// if no good matches were match
	*nOffset	= nBestOffset;
	*nLen		= nBestLen;

	// Update our data stream pointer if we had a good match
	if (*nOffset != 0)
		m_nDataStreamPos = m_nDataStreamPos + (ULONG)(*nLen);

} // FindMatches()


///////////////////////////////////////////////////////////////////////////////
// HashTableInit()
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS_Compress::HashTableInit(void)
{
	int i;
	
	// Clear the hash table
	for (i=0; i<HS_LZSS_HASHTABLE_SIZE; i++)
		m_HashTable[i] = NULL;

	// Clear the hash malloc cache 
	for (i=0; i<HS_LZSS_MALLOC_CACHE; i++)
		m_HashAlloc[i] = NULL;

	m_nHashAllocNumEntries = 0;					// Alloc buffer is not full
}


///////////////////////////////////////////////////////////////////////////////
// HashTableFree()
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS_Compress::HashTableFree(void)
{
	int i;
	struct LZSS_Hash *lpTempNext, *lpTempHash;

	// Free the hash table
	for (i=0; i<HS_LZSS_HASHTABLE_SIZE; i++)
	{
		lpTempHash = m_HashTable[i];
		while (lpTempHash != NULL)
		{
			// Go through the chain freeing each entry
			lpTempNext = lpTempHash->lpNext;
			//free(lpTempHash);
			HashEntryFree(lpTempHash);
			lpTempHash = lpTempNext;
		}
	}

	// Free any alloc buffer entries not taken care of above
	for (i=0; i<HS_LZSS_MALLOC_CACHE; i++)
	{
		if (m_HashAlloc[i] != NULL)
			free(m_HashAlloc[i]);
	}



} // HashTableFree()


///////////////////////////////////////////////////////////////////////////////
// HashTableAdd()
// Adds a hash entry of "nPos" at index "nHash"
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS_Compress::HashTableAdd(UINT nHash, ULONG nPos, ULONG nTooOldPos)
{
	UINT				nTempUINT;
	struct LZSS_Hash	*lpTempNext, *lpTempHash, *lpOldestHash;
	ULONG				nOldestPos;
	ULONG				nTempPos;

	lpTempHash = m_HashTable[nHash];			// Get our first entry at this hash index

	if (lpTempHash == NULL)
	{
		// Initial entry
		//lpTempHash = (_LZSS_Hash*)malloc(sizeof(_LZSS_Hash));
		lpTempHash			= HashEntryAlloc();
		lpTempHash->lpNext	= NULL;
		lpTempHash->nPos	= nPos;
		m_HashTable[nHash]	= lpTempHash;
		return;
	}


	// Existing entry, traverse the chain until we find an empty entry or 
	// exceed the size limit
	nOldestPos = nPos;							// At first, oldest is current!
	nTempUINT = m_nHashChainLimit;
	while (nTempUINT > 0)
	{
		nTempUINT--;

		nTempPos = lpTempHash->nPos;

		// If this entry is too old, replace it
		if (nTempPos < nTooOldPos)
		{
			lpTempHash->nPos = nPos;
			return;
		}

		// Update our oldest hash in case we need to replace
		if (nTempPos < nOldestPos)
		{	
			nOldestPos = nTempPos;
			lpOldestHash = lpTempHash;
		}

		lpTempNext = lpTempHash->lpNext;
		if (lpTempNext == NULL)
		{
			// Found an empty entry
			//lpTempNext = (_LZSS_Hash*)malloc(sizeof(_LZSS_Hash));
			lpTempNext			= HashEntryAlloc();
			lpTempHash->lpNext	= lpTempNext;
			lpTempNext->lpNext	= NULL;
			lpTempNext->nPos	= nPos;
			return;
		}

		lpTempHash = lpTempNext;

	} // End while

	//If we are here, then we ran out of room in the chain, if so, replace the last entry
	lpOldestHash->nPos = nPos;

} // HashTableAdd()


///////////////////////////////////////////////////////////////////////////////
// HashEntryAlloc()
// malloc() is really really slow when allocating and freeing lots of tiny
// 4bytes chunks of memory.  This function acts as a buffer to speed things along
///////////////////////////////////////////////////////////////////////////////

struct LZSS_Hash* HS_LZSS_Compress::HashEntryAlloc(void)
{
	struct LZSS_Hash *lpHash;

	if (m_nHashAllocNumEntries == 0)
	{
		// No unused entries, do it the slow way
		return (struct LZSS_Hash*)malloc(sizeof(LZSS_Hash));
	}

	// Use the last entry in the list
	m_nHashAllocNumEntries--;
	lpHash = m_HashAlloc[m_nHashAllocNumEntries];
	m_HashAlloc[m_nHashAllocNumEntries] = NULL;

	return lpHash;

} // HashMalloc


///////////////////////////////////////////////////////////////////////////////
// HashEntryFree()
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS_Compress::HashEntryFree(struct LZSS_Hash *lpHash)
{

	// If there is no room to buffer this memory release just free it :(
	if (m_nHashAllocNumEntries == HS_LZSS_MALLOC_CACHE)
	{
		free(lpHash);
		return;
	}

	// Enter this entry into our buffer
	m_HashAlloc[m_nHashAllocNumEntries] = lpHash;
	m_nHashAllocNumEntries++;


} // HashEntryFree()
